from twitchpy.client import Client
import ssl
import socket
from twitchpy.message import Message
from twitchpy.channel import Channel

class Bot:
    """
    Class that represents a bot
    """

    def __init__(self,oauth_token,client_id,client_secret,username,channels,command_prefix,ready_message=""):
        """
        Parameters:
        oauth_token (str) -- OAuth token to identify the application
        client_id (str) -- Client ID to identify the application
        client_secret (str) -- Client secret to identify the application
        username (str) -- Name of the bot
        channels (list[str]) -- Channels the bot will access
        command_prefix (str) -- Prefix of the commands the bot will recognize
        ready_message (str) -- Message that the bot will send through the chats of the channels it access
        """

        self.__irc_server="irc.chat.twitch.tv"
        self.__irc_port=6697
        self.__client=Client(oauth_token,client_id,client_secret)
        self.__oauth_token=oauth_token
        self.__access_token=self.__client.get_access_token()
        self.username=username
        self.channels=[]

        for channel in channels:
            self.channels.append(self.__client.get_channel(self.__client.get_user_by_name(channel).id))

        self.command_prefix=command_prefix
        self.ready_message=ready_message
        self.custom_checks={}
        self.custom_listeners={}
        self.custom_commands={}
        self.custom_methods_after_commands={}
        self.custom_methods_before_commands={}

    def __send_privmsg(self,channel,text):
        self.__send_command(f"PRIVMSG #{channel} :{text}")

    def __send_command(self,command):
        if "PASS" not in command:
            print(f"< {command}")

        self.irc.send((command+"\r\n").encode())

    def run(self):
        """
        Method for starting the bot
        """

        self.irc=ssl.wrap_socket(socket.socket())
        self.irc.connect((self.__irc_server,self.__irc_port))
        
        self.__send_command(f"PASS {self.__oauth_token}")
        self.__send_command(f"NICK {self.username}")

        for channel in self.channels:
            self.__send_command(f"JOIN #{channel.name}")
            self.__send_privmsg(channel.name,self.ready_message)

        for check in self.custom_checks.values():
            if check()!=True:
                return

        self.__loop()

    def __get_user_from_prefix(self,prefix):
        domain=prefix.split("!")[0]

        if domain.endswith(".tmi.twitch.tv"):
            return domain.replace(".tmi.twitch.tv","")

        if "tmi.twitch.tv" not in domain:
            return domain

        return None

    def __remove_prefix(self,string,prefix):
        if not string.startswith(prefix):
            return string

        return string[len(prefix):]

    def __parse_message(self,received_msg):
        parts=received_msg.split(" ")

        prefix=None
        user=None
        channel=None
        irc_command=None
        irc_args=None
        text=None
        text_command=None
        text_args=None

        if parts[0].startswith(":"):
            prefix=self.__remove_prefix(parts[0],":")
            user=self.__get_user_from_prefix(prefix)
            parts=parts[1:]

        text_start=next(
            (idx for idx,part in enumerate(parts) if part.startswith(":")),
            None
        )

        if text_start is not None:
            text_parts=parts[text_start:]
            text_parts[0]=text_parts[0][1:]
            text=" ".join(text_parts)

            if text_parts[0].startswith(self.command_prefix):
                text_command=self.__remove_prefix(text_parts[0],self.command_prefix)
                text_args=text_parts[1:]

            parts=parts[:text_start]

        irc_command=parts[0]
        irc_args=parts[1:]

        hash_start=next(
            (idx for idx,part in enumerate(irc_args) if part.startswith("#")),
            None
        )

        if hash_start is not None:
            channel=irc_args[hash_start][1:]

        message=Message(prefix=prefix,user=user,channel=channel,irc_command=irc_command,irc_args=irc_args,text=text,text_command=text_command,text_args=text_args)

        return message

    def __handle_message(self,received_msg):
        if len(received_msg)==0:
            return

        message=self.__parse_message(received_msg)
        print(f"[{message.channel}] {message.user}: {message.text}")

        if message.irc_command=="PING":
            self.__send_command("PONG :tmi.twitch.tv")

        for listener in self.custom_listeners.values():
            listener(message)

        if message.irc_command=="PRIVMSG":
            if message.text_command in self.custom_commands:
                for before in self.custom_methods_before_commands.values():
                    before(message)

                self.custom_commands[message.text_command](message)

                for after in self.custom_methods_after_commands.values():
                    after(message)

    def __loop(self):
        while True:
            received_msgs=self.irc.recv(2048).decode()

            for received_msg in received_msgs.split("\r\n"):
                self.__handle_message(received_msg)

    def add_check(self,name,check):
        """
        Method for adding a check to the bot

        Checks are boolean methods that, if they don't return a False value, will prevent the bot from running

        Parameters:
        name (str) -- Check's name
        check (func) -- Method that will act as a check
        """

        self.custom_checks[name]=check

    def add_listener(self,name,listener):
        """
        Method for adding a listener to the bot

        Listeners are methods that continuously wait for a certain event to occur in order to execute an action

        Parameters:
        name (str) -- Listener's name
        listener (func) -- Method that will act as a listener
        """

        self.custom_listeners[name]=listener

    def add_command(self,name,command):
        """
        Method for adding a command to the bot

        Commands are methods that are executed when someone invokes them in a chat

        Parameters:
        name (str) -- Command's name
        command (func) -- Method that will be executed when the command is invoked
        """

        self.custom_commands[name]=command

    def get_channel(self,id):
        """
        Method that returns a channel

        Parameters:
        id (int) -- Channel's ID

        Return:
        Channel -- If the channel exists
        None -- If the channel doesn't exist
        """

        return self.__client.get_channel(id)

    def get_chatters(self,username):
        """
        Method for getting users into a channel chat

        Parameters:
        channel_name (str) -- Channel's name

        Return:
        dict
        """
        
        return self.__client.get_chatters(username)

    def get_follow(self,from_id,to_id):
        """
        Method for obtaining the date when one user followed another

        Parameters:
        from_id (int) -- ID of the user who is following
        to_id (int) -- ID of the user being followed

        Return:
        str -- If the first user follows the second one
        None -- If the first user does not follow the second one
        """

        return self.__client.get_follow(from_id,to_id)

    def get_followers(self,user_id):
        """
        Method for obtaining a user's followers

        Parameters:
        user_id (int) -- User's ID

        Return:
        list[User]
        """

        return self.__client.get_followers(user_id)

    def get_following(self,user_id):
        """
        Method to obtain the followings of a user

        Parameters:
        user_id (int) -- User's ID

        Return:
        list[User]
        """

        return self.__client.get_following(user_id)

    def get_game_by_id(self,game_id):
        """
        Method for obtaining a category from its ID

        Parameters:
        game_id -- Category's ID

        Return:
        Game -- If the category exists
        None -- If the category doesn't exists
        """

        return self.__client.get_game_by_id(game_id)

    def get_game_by_name(self,game_name):
        """
        Method for obtaining a category from its name

        Parameters:
        game_name -- Category's name

        Return:
        Game -- If the category exists
        None -- If the category doesn't exists
        """

        return self.__client.get_game_by_name(game_name)

    def get_stream_by_channel_id(self,id):
        """
        Method for obtaining a stream from its channel's ID

        Parameters:
        id -- Channel's ID

        Return:
        Stream -- If the channel is live
        None -- If the channel is not live
        """

        return self.__client.get_stream_by_channel_id(id)

    def get_stream_by_username(self,username):
        """
        Method for obtaining a stream from its channel's name

        Parameters:
        username -- Channel's name

        Return:
        Stream -- If the stream is live
        None -- If the stream is not live
        """

        return self.__client.get_stream_by_username(username)

    def get_top_games(self,count=20):
        """
        Get the most viewed categories in Twitch

        Parameters:
        count (int) -- Number of categories to return

        Return:
        list[Game]
        """

        return self.__client.get_top_games(count)

    def get_user_by_id(self,id):
        """
        Method for obtaining a user from his ID

        Parameters: 
        id (int) -- User's ID

        Return:
        User -- If the user exists
        None -- If the user doesn't exist
        """

        return self.__client.get_user_by_id(id)

    def get_user_by_name(self,username):
        """
        Method for obtaining a user from his name

        Parameters: 
        username (str) -- User's name

        Return:
        User -- If the user exists
        None -- If the user doesn't exist
        """

        return self.__client.get_user_by_name(username)

    def get_cheermotes(self,channel_id=None):
        """
        Method for obtaining emotes

        Parameters:
        channel_id (int) -- ID of a channel

        Return:
        dict
        """

        return self.__client.get_cheermotes(channel_id)

    def get_clips_by_channel_id(self,channel_id):
        """
        Method for obtaining clips from a channel

        Parameters:
        channel_id (int) -- Channel's ID

        Return:
        list[dict]
        """

        return self.__client.get_user_by_name(channel_id)

    def get_clips_by_game_id(self,game_id):
        """
        Method for obtaining clips from a category

        Parameters:
        game_id (int) -- Category's ID

        Return:
        list[dict]
        """

        return self.__client.get_clips_by_game_id(game_id)

    def get_clips_by_clip_id(self,clip_id):
        """
        Method for obtaining a clip from its ID

        Parameters:
        clip_id (int) -- Clip's ID

        Return:
        dict -- If the clip exists
        None -- If the clip doesn't exist
        """

        return self.__client.get_clips_by_clip_id(clip_id)

    def get_hype_train_events(self,channel_id):
        """
        Method for obtaining hype train events

        Parameters:
        channel_id (int) -- Channel's ID

        Return:
        list[str] -- If the channel exists
        None -- If the channel doesn't exist
        """

        return self.__client.get_hype_train_events(channel_id)

    def get_streams_by_game_id(self,game_id,count=20):
        """
        Method for obtaining streams from the ID of a category

        Parameters:
        game_id (int) -- Category's ID
        count (int) -- Number of streams to obtain

        Return:
        list[dict] -- If the category exists
        None -- If the category doesn't exist
        """

        return self.__client.get_streams_by_game_id(game_id,count)

    def get_streams_by_language(self,language,count=20):
        """
        Method for obtaining streams in a language

        Parameters:
        language (str) -- Streams' language
        count (int) -- Number of streams to obtain

        Return:
        list[dict] -- If the language exists
        None -- If the language doesn't exist
        """

        return self.__client.get_streams_by_language(language,count)

    def get_stream_tags(self,channel_id=None,tag_id=None,count=20):
        """
        Method for obtaining the tags of a stream

        Parameters:
        channel_id (int) -- Channel's ID
        tag_id (int) -- ID of a tag
        count (int) -- Number of streams to obtain

        Return:
        list[dict] -- If the channel and the tag exist
        None -- If the channel or the tag doesn't exist
        """

        return self.__client.get_stream_tags(channel_id,tag_id,count)

    def get_videos_by_id(self,id):
        """
        Method for obtaining a video from its ID

        Parameters:
        id (int) -- Video's ID

        Return:
        dict -- If the video exists
        None -- If the video doesn't exist
        """

        return self.__client.get_videos_by_id(id)

    def get_videos_by_user_id(self,user_id):
        """
        Method for obtaining the videos of a channel

        Parameters:
        user_id (int) -- Channel's ID

        Return:
        list[dict] -- If the channel exists
        None -- If the channel doesn't exist
        """

        return self.__client.get_videos_by_user_id(user_id)

    def get_videos_by_game_id(self,game_id):
        """
        Method for obtaining the videos of a category

        Parameters:
        game_id (int) -- Category's ID

        Return:
        list[dict] -- If the category exists
        None -- If the category doesn't exist
        """

        return self.__client.get_videos_by_game_id(game_id)

    def add_method_after_commands(self,name,method):
        """
        Method to add to the bot methods that will be executed after each command

        Parameters:
        name (str) -- Method's name
        method (func) -- Method to be executed after each command
        """

        self.custom_methods_after_commands[name]=method

    def add_method_before_commands(self,name,method):
        """
        Method to add to the bot methods that will be executed before each command

        Parameters:
        name (str) -- Method's name
        method (func) -- Method to be executed before each command
        """

        self.custom_methods_before_commands[name]=method

    def remove_check(self,name):
        """
        Method for removing a check from the bot

        Parameters:
        name (str) -- Check's name
        """

        self.custom_checks.pop(name,None)

    def remove_listener(self,name):
        """
        Method for removing a listener from the bot

        Parameters:
        name (str) -- Listener's name
        """

        self.custom_listeners.pop(name,None)

    def remove_method_after_commands(self,name):
        """
        Method to remove a method that is executed after each command

        Parameters:
        name (str) -- Method's name
        """

        self.custom_methods_after_commands.pop(name,None)

    def remove_method_before_commands(self,name):
        """
        Method to remove a method that is executed before each command

        Parameters:
        name (str) -- Method's name
        """

        self.custom_methods_before_commands.pop(name,None)


class Stream:
    """
    Class that represents a stream
    """

    def __init__(self,id,user_id,user_name,game_id,type,title,viewer_count,started_at,language,thumbnail_url,tag_ids):
        """
        Parameters:
        id (int) -- Stream's ID
        user_id (int) -- Channel's ID
        user_name (str) -- Channel's name
        game_id (int) -- Stream's category's ID
        type (str) -- Stream's status
        title (str) -- Stream's title
        viewer_count (int) -- Number of viewers
        started_at (str) -- Stream's start date and time
        language (str) -- Stream's language
        thumbnail_url (str) -- URL of the preview image
        tag_ids (list[int]) -- IDs of the stream's tags
        """

        self.id=id
        self.user_id=user_id
        self.user_name=user_name
        self.game_id=game_id
        self.type=type
        self.title=title
        self.viewer_count=viewer_count
        self.started_at=started_at
        self.language=language
        self.thumbnail_url=thumbnail_url
        self.tag_ids=tag_ids


class User:
    """
    Class that represents an user
    """

    def __init__(self,id,login,display_name,type,broadcaster_type,description,profile_image_url,offline_image_url,view_count):
        """
        Parameters:
        id (int) -- User's ID
        login (str) -- User's login
        display_name (str) -- User's name
        type (str) -- User type
        broadcaster_type (str) -- User's range
        description (str) -- User's description
        profile_image_url (str) -- URL of the user's profile image
        offline_image_url (str) -- URL of the image that is displayed when the user is not on stream
        view_count (int) -- Number of user viewers
        """

        self.id=id
        self.login=login
        self.display_name=display_name
        self.type=type
        self.broadcaster_type=broadcaster_type
        self.description=description
        self.profile_image_url=profile_image_url
        self.offline_image_url=offline_image_url
        self.view_count=view_count


import ssl
import socket
import requests

class Channel():
    """
    Class that represents a channel
    """

    def __init__(self,oauth_token,client_id,client_secret,name,game_name,broadcaster_language,title):
        """
        Parameters:
        oauth_token (str) -- OAuth token to identify the application
        client_id (str) -- Client ID to identify the application
        client_secret (str) -- Client secret to identify the application
        name (str) -- Channel's name
        game_name (str) -- Channel's category's name
        broadcaster_language (str) -- Channel's language
        title -- Title in the channel
        """

        self.__oauth_token=oauth_token
        self.__client_id=client_id
        self.__client_secret=client_secret
        self.__access_token=self.__get_access_token()
        self.__irc_server="irc.chat.twitch.tv"
        self.__irc_port=6697
        self.name=name
        self.game_name=game_name
        self.broadcaster_language=broadcaster_language
        self.title=title
        self.__connect()

    def __get_access_token(self):
        url="https://id.twitch.tv/oauth2/token"
        payload={"client_id":self.__client_id,"client_secret":self.__client_secret,"grant_type":"client_credentials"}

        response=requests.post(url,json=payload).json()

        return response["access_token"]

    def __connect(self):
        self.irc=ssl.wrap_socket(socket.socket())
        self.irc.connect((self.__irc_server,self.__irc_port))
        
        self.__send_command(f"PASS {self.__oauth_token}")
        self.__send_command(f"NICK {self.name}")

        self.__send_command(f"JOIN #{self.name}")

    def __send_command(self,command):
        self.irc.send((command+"\r\n").encode())

    def __send_privmsg(self,channel,text):
        self.__send_command(f"PRIVMSG #{channel} :{text}")

    def ban(self,user,reason=""):
        """
        Method for banning a user

        Parameters:
        user (str) -- User to ban
        reason (str) -- Reason of the ban
        """

        self.__send_privmsg(self.name,f"/ban @{user} {reason}")

    def block(self,user):
        """
        Method for blocking a user

        Parameters:
        user (str) -- User to block
        """

        self.__send_privmsg(self.name,f"/block @{user}")

    def clear(self):
        """
        Method for clearing the chat
        """

        self.__send_privmsg(self.name,"/clear")

    def color(self,color):
        """
        Method to change the color of the channel's name in the chat

        Parameters:
        color (str) -- New color's name
        """

        self.__send_privmsg(self.name,f"/color {color}")

    def commercial(self,duration=30):
        """
        Method for placing advertising in the channel

        Parameters:
        duration (int) -- Duration of advertising
        """

        self.__send_privmsg(self.name,f"/commercial {duration}")

    def emoteonly(self):
        """
        Method for activating the "emotes only" mode
        """

        self.__send_privmsg(self.name,"/emoteonly")

    def emoteonly_off(self):
        """
        Method for disabling "emotes only" mode
        """

        self.__send_privmsg(self.name,"/emoteonlyoff")

    def followers(self):
        """
        Method for activating the "followers only" mode
        """

        self.__send_privmsg(self.name,"/followers")

    def followers_off(self):
        """
        Method for disabling the "followers only" mode
        """

        self.__send_privmsg(self.name,"/followersoff")

    def get_stream(self):
        """
        Method for obtaining information about the channel's stream

        Return:
        Stream -- If the stream is live
        None -- If the stream isn't live
        """

        url=f"https://api.twitch.tv/helix/streams"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.__client_id}
        params={"user_login":self.name}

        response=requests.get(url,headers=headers,params=params).json()
        
        if len(response["data"])>0:
            stream=response["data"][0]
            stream=(stream["id"],stream["user_id"],stream["user_name"],stream["game_id"],stream["type"],stream["title"],stream["viewer_count"],stream["started_at"],stream["language"],stream["thumbnail_url"],stream["tag_ids"],)

            return stream

        else:
            return None

    def host(self,username):
        """
        Method for hosting another channel

        Parameters:
        username (str) -- Name of the channel to host
        """

        self.__send_privmsg(self.name,f"/host {username}")

    def marker(self,description=""):
        """
        Method for leaving a mark on the channel's stream

        Parameters:
        description (str) -- Mark's description
        """

        self.__send_privmsg(self.name,f"/marker {description}")

    def mod(self,username):
        """
        Method for making a user mod

        Parameters:
        username (str) -- Name of the user to be promoted
        """

        self.__send_privmsg(self.name,f"/mod {username}")

    def raid(self,username):
        """
        Method for raiding another channel

        Parameters:
        username -- Name of the channel to raid
        """

        self.__send_privmsg(self.name,f"/raid {username}")

    def send(self,text):
        """
        Method for sending a message by chat

        Parameters:
        text (str) -- Message's text
        """

        self.__send_privmsg(self.name,text)

    def send_me(self,text):
        """
        Method to send a message by chat with the color of the channel's name

        Parameters:
        text (str) -- Message' text
        """

        self.__send_privmsg(self.name,f"/me {text}")

    def slow(self,duration):
        """
        Method for activating the "slow" mode

        Parameters:
        duration (int) -- Time between messages
        """

        self.__send_privmsg(self.name,f"/slow {duration}")

    def slow_off(self):
        """
        Method for disabling the "slow" mode
        """

        self.__send_privmsg(self.name,f"/slow_off")

    def subscribers(self):
        """
        Method for activating the "subscribers only" mode
        """

        self.__send_privmsg(self.name,"/subscribers")

    def subscribers_off(self):
        """
        Method for disabling "subscriber only" mode
        """

        self.__send_privmsg(self.name,"/subscribersoff")

    def timeout(self,user,duration=600,reason=""):
        """
        Method for temporarily expelling a user

        Parameters:
        user (str) -- Name of the user to expel
        duration (int) -- Ejecting time
        reason (str) -- Reason for expulsion
        """

        self.__send_privmsg(self.name,f"/timeout @{user} {duration} {reason}")

    def unban(self,user):
        """
        Method for undoing the ban of a user

        Parameters:
        user (str) -- Name of the user to readmit
        """

        self.__send_privmsg(self.name,f"/unban @{user}")

    def unblock(self,user):
        """
        Method for unblocking a user

        Parameters:
        user (str) -- Name of the user to unblock
        """

        self.__send_privmsg(self.name,f"/unblock @{user}")

    def uniquechat(self):
        """
        Method for activating the "unique" mode
        """

        self.__send_privmsg(self.name,"/uniquechat")

    def uniquechat_off(self):
        """
        Method for disabling the "unique" mode
        """

        self.__send_privmsg(self.name,"/uniquechatoff")

    def unhost(self):
        """
        Method for unhosting the hosted channel
        """

        self.__send_privmsg(self.name,f"/unhost")

    def unmod(self,username):
        """
        Method to remove the moderator's rank from a user

        Parameters:
        username (str) -- User's name
        """

        self.__send_privmsg(self.name,f"/unmod {username}")

    def unraid(self):
        """
        Method to cancel an raid
        """

        self.__send_privmsg(self.name,f"/unraid")

    def unvip(self,username):
        """
        Method to remove the vip range from a user

        Parameters:
        username (str) -- User's name
        """

        self.__send_privmsg(self.name,f"/unvip {username}")

    def vip(self,username):
        """
        Method for making a user vip

        Parameters:
        username (str) -- User's name
        """

        self.__send_privmsg(self.name,f"/vip {username}")

    def whisper(self,user,text):
        """
        Method for whispering to a user

        Parameters:
        user (str) -- User's name
        text (str) -- Whisper's text
        """

        self.__send_privmsg(self.name,f"/w {user} {text}")


class Game:
    """
    Class that represents a Twitch category
    """

    def __init__(self,id,name,box_art_url):
        """
        Parameters:
        id (int) -- Category's ID
        name (str) -- Category's name
        box_art_url (str) -- URL of the category's image
        """

        self.id=id
        self.name=name
        self.box_art_url=box_art_url


class Message:
    """
    Class that represents a message
    """

    def __init__(self,prefix,user,channel,irc_command,irc_args,text,text_command,text_args):
        """
        Parameters:
        prefix (str) -- Message's refix
        user (str) -- User who has sent the message
        channel (str) -- Channel on which the message was sent
        irc_command (str) -- IRC command related to the message
        irc_args (str) -- IRC command's arguments
        text (str) -- Message's text
        text_command (str) -- Command related to the message
        text_args (str) -- Command's arguments
        """

        self.prefix=prefix
        self.user=user
        self.channel=channel
        self.irc_command=irc_command
        self.irc_args=irc_args
        self.text=text
        self.text_command=text_command
        self.text_args=text_args


from setuptools import setup

setup(
    name="TwitchPy",
    version="0.1",
    description="This is a package you can use to connect with the Twitch API, manage a channel, create bots, etc",
    author="DaCasBe",
    author_email="dacasbe97@gmail.com",
    url="https://github.com/DaCasBe/TwitchPy",
    packages=["twitchpy"]
)


import requests
from twitchpy.user import User
from twitchpy.game import Game
from twitchpy.stream import Stream
from twitchpy.channel import Channel

class Client:
    """
    Class that represents a client connection to the Twitch API
    """

    def __init__(self,oauth_token,client_id,client_secret):
        """
        Parámetros:
        oauth_token (str) -- OAuth token to identify the application
        client_id (str) -- Client ID to identify the application
        client_secret (str) -- Client secret to identify the application
        """

        self.oauth_token=oauth_token
        self.client_id=client_id
        self.client_secret=client_secret
        self.__access_token=self.get_access_token()

    def get_access_token(self):
        """
        Method for obtaining a Twitch API access token

        Return:
        str
        """

        url="https://id.twitch.tv/oauth2/token"
        payload={"client_id":self.client_id,"client_secret":self.client_secret,"grant_type":"client_credentials"}

        response=requests.post(url,json=payload).json()

        return response["access_token"]

    def get_chatters(self,channel_name):
        """
        Method for getting users into a channel's chat

        Parameters:
        channel_name (str) -- Channel's name

        Return:
        dict
        """

        url=f"https://tmi.twitch.tv/group/user/{channel_name}/chatters"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}

        response=requests.post(url,headers=headers).json()

        return response["chatters"]

    def get_follow(self,from_id,to_id):
        """
        Method for obtaining the date when one user followed another

        Parameters:
        from_id (int) -- ID of the user who is following
        to_id (int) -- ID of the user being followed

        Return:
        str -- If the first user follows the second one
        None -- If the first user does not follow the second one
        """

        url=f"https://api.twitch.tv/helix/users/follows"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"from_id":from_id,"to_id":to_id}

        response=requests.get(url,headers=headers,params=params).json()
        
        if len(response["data"])>0:
            return response["data"][0]["followed_at"]

        else:
            return None

    def get_followers(self,user_id):
        """
        Method for obtaining a user's followers

        Parameters:
        user_id (int) -- User's ID

        Return:
        list[User]
        """

        url=f"https://api.twitch.tv/helix/users/follows"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"to_id":user_id}

        response=requests.get(url,headers=headers,params=params).json()

        followers=[]

        for follower in response["data"]:
            user=self.get_user_by_id(follower["from_id"])
            followers.append({"follower":user,"followed_at":follower["followed_at"]})

        return followers

    def get_following(self,user_id):
        """
        Method to obtain the followings of a user

        Parameters:
        user_id (int) -- User's ID

        Return:
        list[User]
        """

        url=f"https://api.twitch.tv/helix/users/follows"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"from_id":user_id}

        response=requests.get(url,headers=headers,params=params).json()
        
        followings=[]

        for following in response["data"]:
            user=self.get_user_by_id(following["to_id"])
            followings.append({"follower":user,"followed_at":following["followed_at"]})

        return followings

    def get_game_by_id(self,game_id):
        """
        Method for obtaining a category from its ID

        Parameters:
        game_id -- Category's ID

        Return:
        Game -- If the category exists
        None -- If the category doesn't exists
        """

        url=f"https://api.twitch.tv/helix/games"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"id":game_id}

        response=requests.get(url,headers=headers,params=params).json()
        
        if len(response["data"])>0:
            game=Game(response["data"][0]["id"],response["data"][0]["name"],response["data"][0]["box_art_url"])

            return game

        else:
            return None

    def get_game_by_name(self,game_name):
        """
        Method for obtaining a category from its name

        Parameters:
        game_name -- Category's name

        Return:
        Game -- If the category exists
        None -- If the category doesn't exists
        """

        url=f"https://api.twitch.tv/helix/games"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"name":game_name}

        response=requests.get(url,headers=headers,params=params).json()
        
        if len(response["data"])>0:
            game=Game(response["data"][0]["id"],response["data"][0]["name"],response["data"][0]["box_art_url"])

            return game

        else:
            return None

    def get_stream_by_channel_id(self,id):
        """
        Method for obtaining a stream from its channel's ID

        Parameters:
        id -- Channel's ID

        Return:
        Stream -- If the channel is live
        None -- If the channel is not live
        """

        url=f"https://api.twitch.tv/helix/streams"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"user_id":id}

        response=requests.get(url,headers=headers,params=params).json()
        
        if len(response["data"])>0:
            stream=Stream(response["data"][0]["id"],response["data"][0]["user_id"],response["data"][0]["user_name"],response["data"][0]["game_id"],response["data"][0]["type"],response["data"][0]["title"],response["data"][0]["viewer_count"],response["data"][0]["started_at"],response["data"][0]["language"],response["data"][0]["thumbnail_url"],response["data"][0]["tag_ids"])

            return stream

        else:
            return None

    def get_stream_by_username(self,username):
        """
        Method for obtaining a stream from its channel's name

        Parameters:
        username -- Channel's name

        Return:
        Stream -- If the stream is live
        None -- If the stream is not live
        """

        url=f"https://api.twitch.tv/helix/streams"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"user_login":username}

        response=requests.get(url,headers=headers,params=params).json()
        
        if len(response["data"])>0:
            stream=Stream(response["data"][0]["id"],response["data"][0]["user_id"],response["data"][0]["user_name"],response["data"][0]["game_id"],response["data"][0]["type"],response["data"][0]["title"],response["data"][0]["viewer_count"],response["data"][0]["started_at"],response["data"][0]["language"],response["data"][0]["thumbnail_url"],response["data"][0]["tag_ids"])

            return stream

        else:
            return None

    def get_top_games(self,count=20):
        """
        Get the most viewed categories in Twitch

        Parameters:
        count (int) -- Number of categories to return

        Return:
        list[Game]
        """

        url=f"https://api.twitch.tv/helix/games/top"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"first":count}

        response=requests.get(url,headers=headers,params=params).json()
        games=[]
        
        for game in response["data"]:
            games.append(self.get_game_by_id(game["id"])[0])

        return games

    def get_user_by_id(self,id):
        """
        Method for obtaining a user from his ID

        Parameters: 
        id (int) -- User's ID

        Return:
        User -- If the user exists
        None -- If the user doesn't exist
        """

        url=f"https://api.twitch.tv/helix/users"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"id":id}

        response=requests.get(url,headers=headers,params=params).json()
        
        if len(response["data"])>0:
            user=response["data"][0]
            user=User(user["id"],user["login"],user["display_name"],user["type"],user["broadcaster_type"],user["description"],user["profile_image_url"],user["offline_image_url"],user["view_count"])
            
            return user

        else:
            return None

    def get_user_by_name(self,username):
        """
        Method for obtaining a user from his name

        Parameters: 
        username (str) -- User's name

        Return:
        User -- If the user exists
        None -- If the user doesn't exist
        """

        url=f"https://api.twitch.tv/helix/users"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"login":username}

        response=requests.get(url,headers=headers,params=params).json()
        
        if len(response["data"])>0:
            user=response["data"][0]
            user=User(user["id"],user["login"],user["display_name"],user["type"],user["broadcaster_type"],user["description"],user["profile_image_url"],user["offline_image_url"],user["view_count"])
            
            return user

        else:
            return None

    def get_channel(self,user_id):
        """
        Method that returns a channel

        Parameters:
        id (int) -- Channel's ID

        Return:
        Channel -- If the channel exists
        None -- If the channel doesn't exist
        """

        url=f"https://api.twitch.tv/helix/channels"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"broadcaster_id":user_id}

        response=requests.get(url,headers=headers,params=params).json()
        
        if len(response["data"])>0:
            channel=response["data"][0]
            channel=Channel(self.oauth_token,self.client_id,self.client_secret,self.get_user_by_id(channel["broadcaster_id"]).login,channel["game_name"],channel["broadcaster_language"],channel["title"])
            
            return channel

        else:
            return None

    def get_cheermotes(self,channel_id=None):
        """
        Method for obtaining emotes

        Parameters:
        channel_id (int) -- ID of a channel

        Return:
        dict
        """

        url="https://api.twitch.tv/helix/bits/cheermotes"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}

        if channel_id==None:
            response=requests.get(url,headers=headers).json()

        else:
            params={"broadcaster_id":channel_id}
            response=requests.get(url,headers=headers,params=params).json()

        if len(response["data"])>0:
            return response["data"]

        else:
            return None

    def get_clips_by_channel_id(self,channel_id):
        """
        Method for obtaining clips from a channel

        Parameters:
        channel_id (int) -- Channel's ID

        Return:
        list[dict]
        """

        url="https://api.twitch.tv/helix/clips"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"broadcaster_id":channel_id}

        response=requests.get(url,headers=headers,params=params).json()

        if len(response["data"])>0:
            return response["data"]

        else:
            return None

    def get_clips_by_game_id(self,game_id):
        """
        Method for obtaining clips from a category

        Parameters:
        game_id (int) -- Category's ID

        Return:
        list[dict]
        """

        url="https://api.twitch.tv/helix/clips"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"game_id":game_id}

        response=requests.get(url,headers=headers,params=params).json()

        if len(response["data"])>0:
            return response["data"]

        else:
            return None

    def get_clips_by_clip_id(self,clip_id):
        """
        Method for obtaining a clip from its ID

        Parameters:
        clip_id (int) -- Clip's ID

        Return:
        dict -- If the clip exists
        None -- If the clip doesn't exist
        """

        url="https://api.twitch.tv/helix/clips"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"id":clip_id}

        response=requests.get(url,headers=headers,params=params).json()

        if len(response["data"])>0:
            return response["data"][0]

        else:
            return None

    def get_hype_train_events(self,channel_id):
        """
        Method for obtaining hype train events

        Parameters:
        channel_id (int) -- Channel's ID

        Return:
        list[str] -- If the channel exists
        None -- If the channel doesn't exist
        """

        url="https://api.twitch.tv/helix/hypetrain/events"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"broadcaster_id":channel_id}

        response=requests.get(url,headers=headers,params=params).json()

        if len(response["data"])>0:
            return response["data"]

        else:
            return None

    def get_streams_by_game_id(self,game_id,count=20):
        """
        Method for obtaining streams from the ID of a category

        Parameters:
        game_id (int) -- Category's ID
        count (int) -- Number of streams to obtain

        Return:
        list[dict] -- If the category exists
        None -- If the category doesn't exist
        """

        url="https://api.twitch.tv/helix/streams"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"game_id":game_id,"first":f"{count+1}"}

        response=requests.get(url,headers=headers,params=params).json()

        if len(response["data"])>0:
            return response["data"]

        else:
            return None

    def get_streams_by_language(self,language,count=20):
        """
        Method for obtaining streams in a language

        Parameters:
        language (str) -- Streams' language
        count (int) -- Number of streams to obtain

        Return:
        list[dict] -- If the language exists
        None -- If the language doesn't exist
        """

        url="https://api.twitch.tv/helix/streams"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"language":language,"first":count}

        response=requests.get(url,headers=headers,params=params).json()

        if len(response["data"])>0:
            return response["data"]

        else:
            return None

    def get_stream_tags(self,channel_id=None,tag_id=None,count=20):
        """
        Method for obtaining the tags of a stream

        Parameters:
        channel_id (int) -- Channel's ID
        tag_id (int) -- ID of a tag
        count (int) -- Number of streams to obtain

        Return:
        list[dict] -- If the channel and the tag exist
        None -- If the channel or the tag doesn't exist
        """

        url="https://api.twitch.tv/helix/tags/streams"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}

        if channel_id==None:
            if tag_id==None:
                response=requests.get(url,headers=headers).json()

            else:
                params={"tag_id":tag_id,"first":count}
                response=requests.get(url,headers=headers,params=params).json()

        else:
            if tag_id==None:
                response=requests.get(url,headers=headers).json()

            else:
                params={"broadcaster_id":channel_id,"tag_id":tag_id,"first":count}
                response=requests.get(url,headers=headers,params=params).json()

        if len(response["data"])>0:
            return response["data"]

        else:
            return None

    def get_videos_by_id(self,id):
        """
        Method for obtaining a video from its ID

        Parameters:
        id (int) -- Video's ID

        Return:
        dict -- If the video exists
        None -- If the video doesn't exist
        """

        url="https://api.twitch.tv/helix/videos"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"id":id}

        response=requests.get(url,headers=headers,params=params).json()

        if len(response["data"])>0:
            return response["data"][0]

        else:
            return None

    def get_videos_by_user_id(self,user_id):
        """
        Method for obtaining the videos of a channel

        Parameters:
        user_id (int) -- Channel's ID

        Return:
        list[dict] -- If the channel exists
        None -- If the channel doesn't exist
        """

        url="https://api.twitch.tv/helix/videos"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"user_id":user_id}

        response=requests.get(url,headers=headers,params=params).json()

        if len(response["data"])>0:
            return response["data"]

        else:
            return None

    def get_videos_by_game_id(self,game_id):
        """
        Method for obtaining the videos of a category

        Parameters:
        game_id (int) -- Category's ID

        Return:
        list[dict] -- If the category exists
        None -- If the category doesn't exist
        """

        url="https://api.twitch.tv/helix/videos"
        headers={"Authorization": f"Bearer {self.__access_token}","Client-Id":self.client_id}
        params={"game_id":game_id}

        response=requests.get(url,headers=headers,params=params).json()

        if len(response["data"])>0:
            return response["data"]

        else:
            return None